/* config.h.  Generated from config.h.in by configure.  */
/* config.h.in.  Generated from configure.ac by autoheader.  */

#include "xorg-server.h"

/* Define to 1 if you have the <dlfcn.h> header file. */
#define HAVE_DLFCN_H 1

/* Define to 1 if you have the <inttypes.h> header file. */
#define HAVE_INTTYPES_H 1

/* Define to 1 if you have the <stdint.h> header file. */
#define HAVE_STDINT_H 1

/* Define to 1 if you have the <stdio.h> header file. */
#define HAVE_STDIO_H 1

/* Define to 1 if you have the <stdlib.h> header file. */
#define HAVE_STDLIB_H 1

/* Define to 1 if you have the <strings.h> header file. */
#define HAVE_STRINGS_H 1

/* Define to 1 if you have the <string.h> header file. */
#define HAVE_STRING_H 1

/* Define to 1 if you have the <sys/stat.h> header file. */
#define HAVE_SYS_STAT_H 1

/* Define to 1 if you have the <sys/types.h> header file. */
#define HAVE_SYS_TYPES_H 1

/* Define to 1 if you have the <unistd.h> header file. */
#define HAVE_UNISTD_H 1

/* Define to the sub-directory where libtool stores uninstalled libraries. */
#define LT_OBJDIR ".libs/"

/* Name of package */
#define PACKAGE "xlibre-xf86-video-savage"

/* Define to the address where bug reports for this package should be sent. */
#define PACKAGE_BUGREPORT "https://github.com/X11Libre/xf86-video-savage/issues"

/* Define to the full name of this package. */
#define PACKAGE_NAME "xlibre-xf86-video-savage"

/* Define to the full name and version of this package. */
#define PACKAGE_STRING "xlibre-xf86-video-savage 25.0.0"

/* Define to the one symbol short name of this package. */
#define PACKAGE_TARNAME "xlibre-xf86-video-savage"

/* Define to the home page for this package. */
#define PACKAGE_URL ""

/* Define to the version of this package. */
#define PACKAGE_VERSION "25.0.0"

/* Major version of this package */
#define PACKAGE_VERSION_MAJOR 25

/* Minor version of this package */
#define PACKAGE_VERSION_MINOR 0

/* Patch version of this package */
#define PACKAGE_VERSION_PATCHLEVEL 0

/* Enable DRI driver support */
#define SAVAGEDRI 1

/* Enable developmental DRI driver support */
#define SAVAGEDRI_DEVEL 1

/* Define to 1 if all of the C89 standard headers exist (not just the ones
   required in a freestanding environment). This macro is provided for
   backward compatibility; new code need not use it. */
#define STDC_HEADERS 1

/* Version number of package */
#define VERSION "25.0.0"
