/*
 * Copyright 2011 X.Org Foundation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * COPYRIGHT HOLDER(S) OR AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Except as contained in this notice, the name of the copyright holder(s)
 * and author(s) shall not be used in advertising or otherwise to promote
 * the sale, use or other dealings in this Software without prior written
 * authorization from the copyright holder(s) and author(s).
 */

#ifndef SAVAGE_PCIIDS_H
#define SAVAGE_PCIIDS_H

#define PCI_CHIP_SAVAGE3D		0x8A20
#define PCI_CHIP_SAVAGE3D_MV		0x8A21
#define PCI_CHIP_SAVAGE4		0x8A22
#define PCI_CHIP_PROSAVAGE_PM		0x8A25
#define PCI_CHIP_PROSAVAGE_KM		0x8A26
#define PCI_CHIP_SAVAGE_MX_MV		0x8C10
#define PCI_CHIP_SAVAGE_MX		0x8C11
#define PCI_CHIP_SAVAGE_IX_MV		0x8C12
#define PCI_CHIP_SAVAGE_IX		0x8C13
#define PCI_CHIP_SUPSAV_MX128		0x8C22
#define PCI_CHIP_SUPSAV_MX64		0x8C24
#define PCI_CHIP_SUPSAV_MX64C		0x8C26
#define PCI_CHIP_SUPSAV_IX128SDR	0x8C2A
#define PCI_CHIP_SUPSAV_IX128DDR	0x8C2B
#define PCI_CHIP_SUPSAV_IX64SDR		0x8C2C
#define PCI_CHIP_SUPSAV_IX64DDR		0x8C2D
#define PCI_CHIP_SUPSAV_IXCSDR		0x8C2E
#define PCI_CHIP_SUPSAV_IXCDDR		0x8C2F
#define PCI_CHIP_S3TWISTER_P		0x8D01
#define PCI_CHIP_S3TWISTER_K		0x8D02
#define PCI_CHIP_PROSAVAGE_DDR		0x8D03
#define PCI_CHIP_PROSAVAGE_DDRK		0x8D04
#define PCI_CHIP_SAVAGE2000		0x9102

#endif /* SAVAGE_PCIIDS_H */
